// OSIHFE.cpp : Defines the entry point for the console application.
// Mark Spankus 2018
//
#include "osihfe.h"

#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdarg.h>
#include "hfe.h"


FILE* logfile = stderr;
bool bVerbose = false;

const TCHAR * szIdentity[MAX_TYPE] = {
	_T("Unknown\0") ,
	_T("\0"),
	_T("\0"),
	_T("\0"),
	_T("\0"),
	_T("OS65D5\0"),
	_T("HEXDOS\0"),
	_T("DOS/65-5\0"),
	_T("OS65D8\0"),
	_T("OS65U\0"),
	_T("DOS/65-8\0"),
	_T("Empty 8\"\0"),
	_T("Empty 5.25\"\0"),
	_T("Empty 8\"\0"),
	_T("Empty 5.25\"\0"),
};


void usage()
{
	_fputts( _T("Usage: OSIHFE -[bcdiov5l] infile [outfile]\r\n       Converts between OSI disk dump and HxC floppy emulator file formats\r\n"), stdout);
	_fputts( _T("       If [outfile] is not specified, a target file is automatically\r\n")\
	         _T("          created based on input filetype\r\n"), stdout);
	_fputts( _T("       -b65D5/-b65D8/-b65U/-bHEXDOS -b8/-b8z -b5/-b5z outfile :create blank OSI HFE\r\n          image of specified type or empty (-b8/-b5) or null (-b5z/-b8z)\r\n"), stdout);
	_fputts( _T("       -c infile outfile :clean OSI disk dump image\r\n"), stdout);
	_fputts( _T("       -d infile :list directory contents\r\n"), stdout);
	_fputts( _T("       -i infile :identify type (OS65D5, OS65D8, OS65U, HEXDOS)\r\n"), stdout);
	_fputts( _T("       -v :turn on verbose output\r\n"), stdout);
	_fputts( _T("       -5 :write 8\" image at 5.25\" bit rate to allow mounting 8\" image on 5.25\"\r\n           system\r\n"), stdout);
	_fputts( _T("       -o :overwrite existing output file\r\n"), stdout);
	_fputts( _T("       -l<filename> :specify verbose log output file name instead of stderr\r\n"), stdout);
	_fputts( _T(" (ver 1.4f)\r\n"), stdout);

}


int _tmain(int argc, TCHAR* argv[])
{
	
	FILE* inpfile = NULL;
	FILE* outfile = NULL;
	

	bool bBlank = false;
	bool bClean = false;
	bool bOverwrite = false;
	bool bIdentify = false;
	bool bWrite8As5 = false;
	bool bDirectory = false;

	DISKTYPE dTarget = DISK_UNKNOWN;


	unsigned char* pMem = NULL;
	unsigned char *pEnd = NULL;

	int nSide = 1;
	DISKTYPE dt = DISK_UNKNOWN;
	unsigned char * pOSIDisk = NULL;
	int nOSILen = 0;
	int nReturn = 0;
	
	TCHAR szFile[8192] = {0};



    while (--argc && ((*++argv)[0] == _T('-') )) 
	{
		const TCHAR * arg = argv[0];
		arg++;
		if (!_tcsicmp(arg, _T("v"))) 
		{          
            bVerbose = true;
			continue;
        }

		if (!_tcsicmp(arg, _T("o")))
		{
			bOverwrite = true;
			continue;
		}
		if (!_tcsicmp(arg, _T("i")))
		{
			bIdentify = true;
			continue;
		}
		if (!_tcsicmp(arg, _T("c")))
		{
			bClean = true;
			continue;
		}
		if (!_tcsicmp(arg, _T("d")))
		{
			bDirectory = true;
			continue;
		}
		if (!_tcsicmp(arg, _T("5")))
		{
			bWrite8As5 = true;
			continue;
		}
		if (!_tcsicmp(arg, _T("b65D5")))
		{
			bBlank = true;
			dTarget = OS65D5;
			continue;
		}
		if (!_tcsicmp(arg, _T("b65D8")))
		{
			bBlank = true;
			dTarget = OS65D8;
			continue;
		}
		if (!_tcsicmp(arg, _T("b65U")))
		{
			bBlank = true;
			dTarget= OS65U8;
			continue;
		}
		if (!_tcsicmp(arg, _T("bHEXDOS")))
		{
			bBlank = true;
			dTarget= HEXDOS;
			continue;
		}
		if (!_tcsicmp(arg, _T("b8")))
		{
			bBlank = true;
			dTarget= EMPTY8;
			continue;
		}
		if (!_tcsicmp(arg, _T("b5")))
		{
			bBlank = true;
			dTarget= EMPTY5;
			continue;
		}
		if (!_tcsicmp(arg, _T("b8z")))
		{
			bBlank = true;
			dTarget= EMPTY8Z;
			continue;
		}
		if (!_tcsicmp(arg, _T("b5z")))
		{
			bBlank = true;
			dTarget= EMPTY5Z;
			continue;
		}
		if ( (*arg == _T('l') || *arg == _T('L'))  && _tcslen(arg) > 1) //logfile
		{
			if ( (logfile = _tfopen(arg+1, _T("wb"))) == NULL )
				logfile = stderr;
			continue;

		}



		_fputts( _T(" Unknown argument:"), stderr);
		_fputts( arg, stderr);
		_fputts( _T("\r\n"), stderr);

		//anything else
		usage();
		return 0;


	}	
	
	if (argc > 0)
	{
	    int i = 0;

		_tcsncpy(szFile, argv[i], 8191);

		if (!bBlank)
		{
			if ( (inpfile = _tfopen(argv[i], _T("rb"))) == NULL)
			{
				_ftprintf(stderr, _T("osihfe: Can't open '%s' for input.\n"), argv[i]);
				return -1;
			}
		}
		else
		{
			if (argc != 1)
			{
				_ftprintf(stderr, _T("osihfe: Too many arguments for blank disk creation.\n"));	
				return -1;
			}
			i--;
		}


		if ( ++i < argc )
		{

			if ( (outfile = _tfopen(argv[i], _T("r+b"))) != NULL )
			{
				if (!bOverwrite)
				{
					if (outfile)
					{
						fclose(outfile);
						outfile = NULL;
					}
					if (inpfile)
					{
						fclose(inpfile);
						inpfile = NULL;
					}
					_ftprintf(stderr, _T("osihfe: Target file '%s' exists. Use -o to overwrite\n"), argv[i]);
					return -1;
				}
			}
			else if ((outfile = _tfopen(argv[i], _T("wb"))) == NULL )
			{
				if (inpfile)
					fclose(inpfile);
				_ftprintf(stderr, _T("osihfe: Can't open '%s' for output.\n"), argv[i]);
				return -1;
			}

		}
		else if (bClean)
		{
			fclose(inpfile);
			_ftprintf(stderr, _T("osihfe: Target filename required for clean operation\r\n"));
			return -1;
		}
	}
	else
	{
		usage();
		return 0;
	}
	/*if (outfile == stdout) //ensure our binary \n's don't convert to \r\n!!!!!
	{
		_setmode(fileno(stdout), _O_BINARY);
	}*/


	if (bBlank)
	{
		if (CreateHFEImage( NULL, dTarget, 2, &pOSIDisk, &nOSILen, bWrite8As5 ))
		{
			if (bWrite8As5 && pOSIDisk)
			{
				picfileformatheader* pHead = (picfileformatheader*) pOSIDisk;
				if (pHead->bitRate > 250)
					pHead->bitRate = 250;
				pHead->floppyRPM = 0;
			}
			
			size_t nWrote = fwrite(pOSIDisk, 1, nOSILen, outfile);
			if (nWrote != (size_t) nOSILen)
			{
				LogText(_T("Error writing to target file\r\n") );	
			}
		}


	}
	else
	{

		struct _stat buf;
		int result = _fstat( _fileno(inpfile), &buf );

		size_t nReadBytes = buf.st_size;

		if ( !result && (nReadBytes < 1000 || nReadBytes > 5000000)) 
		{
			//invalid data
			LogText(_T("Invalid HFE file input size"));
			goto HFEabort;

		}
		else if (result)
		{
			//just make a really big buffer for stdin?
			nReadBytes = 4000000;

		}

		pMem = (unsigned char*) ::malloc (nReadBytes);
		if (!pMem)
		{
			LogText( _T("Error Can't allocate %d bytes"),nReadBytes);
			return -1; 
		}
		if (pMem)
		{
			pEnd = pMem + nReadBytes;

			//read data into memory
			size_t nRead = fread( pMem, sizeof( char ), nReadBytes, inpfile );

			if (nRead != nReadBytes)
			{
				if ( ferror(inpfile))
				{
					//error reading from file
					LogText( _T("Error reading from input file\r\n"));
					fclose(inpfile);

					goto HFEabort;
				}
				else if ( feof(inpfile))
				{
					nReadBytes = nRead;	
				}
			}

			fclose(inpfile); //done reading


			if (memcmp(pMem, szHeaderSig, 8))	
			{
				//looks like NOT HFE file
				dt = IdentifyDiskType( pMem, nReadBytes);

				
				if (bIdentify)
				{
					LogText( _T("File type %s\r\n"), szIdentity[(int) dt]);
					nReturn = dt;
					
				}
				else if (bDirectory)
				{
					TCHAR *pText = NULL;
					pText = GetOSIDirectory(pMem, dt);
					
					if (pText)
					{
						_fputts(pText, stdout);
						free(pText);
						pText = NULL;
					}
				}
				else
				{
					if (bClean)
					{
						if (CleanOSIImage(pMem, (DISKTYPE) dt,  &pOSIDisk, &nOSILen ))
						{

							if (outfile)
							{
								size_t nWrote = fwrite(pOSIDisk, 1, nOSILen, outfile);
								if (nWrote != (size_t) nOSILen)
								{
									LogText(_T("Error writing to target file\r\n") );	
								}
							}
							else
							{
								_ftprintf(stderr, _T("osihfe: You must specify a target file for Clean\r\n") );
							}


						}
						else
						{
							_ftprintf(stderr, _T("osihfe: Can't clean this disk file.") );
						}
					}
					else
					{
						if (CreateHFEImage( pMem, (DISKTYPE) dt, 2, &pOSIDisk, &nOSILen, bWrite8As5 ))
						{
							if (bWrite8As5 && pOSIDisk)
							{
								picfileformatheader* pHead = (picfileformatheader*) pOSIDisk;
								if (pHead->bitRate > 250)
									pHead->bitRate = 250;
								pHead->floppyRPM = 0;
							}
							
							if (!outfile)
							{
								//make an output file base on input now that we know what file
								//we are making
								int nLen = _tcslen(szFile);
								int j = nLen-1;
								while (	j > 0 && szFile[j] != _T('.'))
									j--;
								if (szFile[j] == _T('.'))
									szFile[j] = 0;	
								_tcscat(szFile, _T(".HFE"));

								if ( (outfile = _tfopen(szFile, _T("r+b"))) != NULL )
								{
									if (!bOverwrite)
									{
										fclose(outfile);
										fclose(inpfile);
										_ftprintf(stderr, _T("osihfe: Target file '%s' exists. Use -o to overwrite\n"), szFile);
										goto HFEabort;
									}
								}
								else if ((outfile = _tfopen(szFile, _T("wb"))) == NULL )
								{
									fclose(inpfile);
									_ftprintf(stderr, _T("osihfe: Can't open '%s' for output.\n"), szFile);
									goto HFEabort;
								}

									

							}
							
							size_t nWrote = fwrite(pOSIDisk, 1, nOSILen, outfile);
							if (nWrote != (size_t) nOSILen)
							{
								LogText(_T("Error writing to target file\r\n") );	
							}
						}
					}
				}


			}
			else
			{
				//found FE header...
				
				//decode HFE src to figure out target type, could read #tracks too
				dt = DetermineOSIDiskTypeFromHFE(pMem, nReadBytes, nSide );
				if (dt <= 0)
				{
					LogText(_T("Unable to determine target format from HFE file\r\n") );
					goto HFEabort;	
				}
				if (bIdentify)
				{
					LogText( _T("File type %s\r\n"), szIdentity[(int) dt]);
					nReturn = dt;
					goto HFEabort;
				}


				if (CreateOSIImage( pMem, nReadBytes, nSide, &pOSIDisk, &nOSILen , bVerbose ))
				{

					if (bDirectory)
					{
						
						if (pOSIDisk)
						{
							TCHAR *pText = NULL;
							pText = GetOSIDirectory(pOSIDisk, dt);

							if (pText)
							{
								_fputts(pText, stdout);
								free(pText);
								pText = NULL;
							}
						}
					}
					else
					{
							
						if (!outfile)
						{
							//make an output file base on input now that we know what file
							//we are making
							int nLen = _tcslen(szFile);
							int j = nLen-1;
							while (	j > 0 && szFile[j] != _T('.'))
								j--;
							if (szFile[j] == _T('.'))
								szFile[j] = 0;
							switch (dt)
							{
								case OS65D5:
								case OS65D8:
								default:									
									_tcscat(szFile, _T(".65D"));
									break;
								case OS65U8:
									_tcscat(szFile, _T(".65U"));
									break;
							}
							
							if ( (outfile = _tfopen(szFile, _T("r+b"))) != NULL )
							{
								if (!bOverwrite)
								{
									fclose(outfile);
									fclose(inpfile);
									_ftprintf(stderr, _T("osihfe: Target file '%s' exists. Use -o to overwrite\n"), szFile);
									goto HFEabort;
								}
							}
							else if ((outfile = _tfopen(szFile, _T("wb"))) == NULL )
							{
								fclose(inpfile);
								_ftprintf(stderr, _T("osihfe: Can't open '%s' for output.\n"), szFile);
								goto HFEabort;
							}
						}

					
						size_t nWrote = fwrite(pOSIDisk, 1, nOSILen, outfile);
						if (nWrote != (size_t) nOSILen)
						{
							LogText(_T("Error writing to target file\r\n") );	
						}
					}
				}
			}
		}
	}

HFEabort:

	if (pOSIDisk)
	{
		free(pOSIDisk);
		pOSIDisk = NULL;
	}

	if (pMem)
	{
		free(pMem);
		pMem = NULL;
	}
	if (outfile)
		fclose(outfile);
	if (logfile != NULL && logfile != stderr)
		fclose(logfile);
	
	return nReturn;
}
static int logpos = 0;
void LogData(unsigned char c)
{
	TCHAR hex[17] = _T("0123456789ABCDEF");
	TCHAR out[4] = {0};
    if (logfile)
	{
		
		if ((logpos%16) == 0)
		{
			_fputts( _T("\r\n"), logfile);
			logpos = 0;
		}
		logpos++;

		out[0] = hex[c >> 4];
		out[1] = hex[c &0x0F];
		out[2] = 0x20;
		out[3] = 0;
		_fputts( out, logfile);
	}
}
void LogText(const TCHAR * szFormat, ...)
{
	static TCHAR buf[2048];
    va_list vargs;

    if (logfile)
	{
		if (szFormat && *szFormat) 
		{
			va_start( vargs, szFormat );
			_vsntprintf( buf, 2047, szFormat, vargs );
			va_end( vargs );  
        
			_fputts( buf, logfile);
			logpos = 0;
		}
	}
}
